P2 TACHYON
IFDEF VOCAB:
VOCAB: ETHERNET
ETHERNET
}

IFDEF W5500.fth
W5500.dat DATORG
FORGET W5500.fth
}

pub W5500.fth		PRINT" WIZNET W5500 driver for TF2 171217.0000 " ;

DATPTR	== W5500.dat

IFNDEF EE!
ALIAS ! EE!
ALIAS @ EE@
}

: PCB$	" P2DEV" ;

--- some default IP settings that are only loaded in on a new pcb and if storage locations in upper eeprom is blank
&192.168.0.1		== myGW
&192.168.0.151		== myIP
&255.255.255.0		== mySN



( HARDWARE DEFINITIONS )


{ W5500 SPI frame format
ADDRH ADDRL CONTROL <DATA>
7..5	SKT
4,3	X,REG,TX,RX
2	R/W
1,0	00 = VARIABLE LEN DATA
}

&20.23.22.21	== @wizpins
pub WIZPINS	' @wizpins 2+ ! ;

--- Write and update CONTROL byte
pri RDWIZ						0
---							ctl byte
pri WIZADR ( wr -- )					SWAP @wizpins SPIPINS SPIWR16
--- entry point only for reference by SWC!33
pri SWC							0_0 OR SPIWR8 DROP ;
pri WRWIZ						4 WIZADR ;

pub @COMMON ( 6.8us )					0 $F8
--- write directly to the control byte literal embedded in SWC
pri SWC! ( or andn -- )					' SWC C@ SWAP ANDN OR ' SWC C! ;
pub @RX							$18 $18 SWC! ;
pub @TX							$10 $18 SWC! ;
pub @SOCKET						8 $18 SWC! ;

\ BYTE skt1
BYTE skt2
---							socket constant
pub SKT@						0_0 ;
---		Set the socket number ( calculates address and sets as a SOCKET constant )
pub SOCKET ( skt -- )					DUP ' SKT@ C! 5 << $E0 SWC! ;

ALIAS SOCKET SKT

--- READ/WRITE METHODS ---
pub LC! ( byte wizadr -- ) ( 17.5us )			WRWIZ SPIWR8 DROP SPICE ;
pub LW! ( word wizadr -- ) ( 18.8us )			WRWIZ SPIWR16 SPICE ;
pub L! ( long wizadr -- ) ( 21.1us )			WRWIZ SPIWR SPIWR SPIWR SPIWR DROP SPICE ;

pub LC@ ( wizadr -- byte ) ( 16.8us )			RDWIZ 0 SPIRD SPICE ;
pub LW@ ( wizadr -- word ) ( 18.0us )			RDWIZ 0 SPIRD SPIRD SPICE ;
pub L@ ( wizadr -- long ) ( 20.3us )			RDWIZ 0 SPIRD SPIRD SPIRD SPIRD SPICE ;

--- BLOCK TRANSFER METHODS ---
--- Read a block of memory from the WIZnet
pub LREAD ( src dst cnt -- )				ROT RDWIZ SPI>BUF SPICE ;

--- Write a block to the WIZnet - 558us/512 bytes
pub LWRITE ( src dst cnt -- )				SWAP WRWIZ BUF>SPI SPICE ;



( ----------------------------------- REGISTER LEVEL INTERFACE -------------------------------- )

--- EEPROM CONFIG BACKUP ---
--- Use the top part of the 64K EEPROM to save IP settings
DATPTR
$FFC0	DATORG
4	BYTES @gateway
4	BYTES @sip
4	BYTES @subnet
4	BYTES @mac
4	BYTES @wcold
#32	BYTES @ports
	DATORG
\ common registers

--- Access common registers
pub wMODE ( mask -- )			0 @COMMON LC! ;
pub GATEWAY ( addr -- )			DUP @gateway EE! 1 @COMMON L! ;
pub SUBNET ( mask -- )			DUP @subnet EE! 5 @COMMON L! ;
pub MAC ( high low  -- )		DUP @mac EE! $0B @COMMON L! 9 LW! ;
pub SIP ( long -- )			DUP @sip EE! $0F @COMMON L! ;


pri INTS@ ( -- ints )			$15 @COMMON LC@ ;
pri INTMASK ( mask -- )			$16 @COMMON LC! ;


pri @RTR ( -- adr )			$19 @COMMON ;
pri @RCR ( -- adr )			$1B @COMMON ;

pri UIP ( ip -- )			$28 @COMMON L! ;
pri UPORT ( port -- )			$2C @COMMON LW! ;



( SOCKET INTERRUPTS )
4	|< == 	&SENDOK
3	|< == 	&TIMEOUT
2	|< == 	&RECV
1	|< == 	&DISCON
0	|< == 	&CON
{
( SOCKET STATUS CODES )
$00	==	SOCK_CLOSED
$13	==	SOCK_INIT
$14	==	SOCK_LISTEN
$17	==	SOCK_ESTABLISHED
$1C	==	SOCK_CLOSE_WAIT
$22	==	SOCK_UDP
$32	==	SOCK_IPRAW
$42	==	SOCK_MACRAW
$5F	==	SOCK_PPOE

( SOCKET TRANSIENT STATUS CODES )
$15	==	SOCK_SYNSENT
$16	==	SOCK_SYNRECV
$18	==	SOCK_FIN_WAIT
$1A	==	SOCK_CLOSING
$1B	==	SOCK_TIME_WAIT
$1D	==	SOCK_LAST_ACK
$11	==	SOCK_ARP
$21	==	SOCK_ARP1
$31	==	SOCK_ARP2
}


\ Socket registers ( -- addr )
pri sMODE		0 @SOCKET ;		--- mode
pri sCMD!		1 @SOCKET LC! ;		--- command
pri sINTS		2 @SOCKET ;		--- interrupts
pri sSTAT@		3 @SOCKET LC@ ;		--- status
pri sPORT		4 @SOCKET ;		--- 2 byte src port
pri sDHAR		6 @SOCKET ;		--- 6 byte dest hardware address
pri sDIP		$0C @SOCKET ;		--- 4 byte dest IP address
pri sDPORT		$10 @SOCKET ;		--- 2 byte dest port
pri sSSIZE		$12 @SOCKET ;		--- 2 byte dest max seg size
pri sPRO		$14 @SOCKET ;		--- protocol in IP raw mode

pri sRXMEM		$1E @SOCKET ;
pri sTXMEM		$1F @SOCKET ;
pri TXFREE@		$20 @SOCKET LW@ ;
pri TXREAD		$22 @SOCKET ;
pri TXWRITE		$24 @SOCKET ;
pri RXSIZE@		$26 @SOCKET LW@ ;
pri RXREAD		$28 @SOCKET ;
pri RXWRITE		$2A @SOCKET ;
pri KEEPTMR		$2D @SOCKET ;	--- keep alive timer

\ Protocol modes
pri CLOSED		0 sMODE LC! ;
pub TCP			1 sMODE LC! ;
pub UDP			2 sMODE LC! ;
pri IPRAW		3 sMODE LC! ;
pri MACRAW		4 sMODE LC! ;
pri PPPoE		5 sMODE LC! ;





( SOCKET COMMANDS - use currently select socket )

pub sOPEN		1 sCMD! ;
pub sLISTEN		2 sCMD! ;
pub sCONNECT		4 sCMD! ;
pub sDISCON		8 sCMD! ;
pub sCLOSE		$10 sCMD! ;
pub sSEND		$20 sCMD! ;
pub sSENDMAC		$21 sCMD! ;
pub sSENDKEEP		$22 sCMD! ;
pub sRECV		$40 sCMD! ;


pri sCLOSED? 		sSTAT@ 0= ;
pri sINACTIVE?		sSTAT@ $17 < ;
pri sESTAB?		sSTAT@ $17 = ;
pri sCLOSING?		sSTAT@ $18 $1B WITHIN ;	--- closing or wait closing
pri sCONNECTED?		sINTS LC@ &CON AND DUP IF &CON sINTS LC! THEN ; --- Test and reset the connection interrupt;
pri sDISCON?		sINTS LC@ &DISCON AND DUP IF -1 sINTS LC! THEN ;





pub SetPORT
 	sPORT LW!
 	;
pub PORT! ( srcport -- ) --- save port config and set port
 	DUP SKT@ 2* 2* @ports + EE!
	SetPORT
	;

( Set the factory defaults on very first run - generate a random MAC in the 01.FF.xx.xx.xx.xx range )
pub WCOLD
	CR PRINT" Setting default IP configuration "
 	myGW @gateway EE!
 	myIP @sip EE!
 	mySN @subnet EE!
 	RND @mac EE! 				\ random MAC at first run, Need to have a unique ID
 	8 0 DO 0 I 2* 2* @ports + EE! LOOP
 	$A55A @wcold EE!
 	;

\ Init the SPI for the WIZnet chip
pub !WIZIO
\	!PCB
	@wizpins SPIPINS				\ Use SPI
	@COMMON
\	ON WRESET OFF WRESET
	;

$02FF == oui2

pub !WIZIP
 	@wcold EE@ $A55A <> IF WCOLD THEN
	@gateway EE@ GATEWAY			\ Assign a default gateway
	@subnet EE@ SUBNET
	@sip EE@ SIP
	oui2 @mac EE@ MAC
  	8 0 DO I 2* 2* @ports + EE@ I SOCKET SetPORT LOOP
	;

pri !TXBUFS	8 FOR I SOCKET BUFFERS $800 ERASE @TX BUFFERS 0 $800 LWRITE @RX BUFFERS 0 $800 LWRITE NEXT ;

pub !WIZ 	!WIZIO !TXBUFS !WIZIP ;


{ Since the W5500??? read and write index registers are unreadable until a connection has been established
or cannot be written incrementally until a send then these are buffered for when they become readable
}

16 BYTES txwr 						--- tx write buffers for 8 sockets

pub @txwr ( -- addr )   	txwr SKT@ 2* + ;

pub !TXWR		txwr 16 $FF FILL ;				--- invalidate all txwr pointers (W5500 workaround)

LONG txtime
LONG txsize

BYTE autosend							--- Flag to control whether LANSEND sends when it receives a LF or not
$0A autosend C!							--- preset to autosend

\ TIMER sendtimer
pri WAITSEND
 	txsize @ 0EXIT 						--- don't bother if it's empty
	#1000
 	FOR
 	  sINTS LC@ &SENDOK AND 				--- until sent
 	  TXREAD LW@ TXWRITE LW@ = AND
 	?NEXT DROP
 	&SENDOK sINTS LC!					--- reset the sent interrupt flag
 	;

\ Request WIZnet to send off current transmit data in buffer
pub LANSEND
 	&SENDOK sINTS LC!					--- Clear send interrupt
 	@txwr W@ TXWRITE LW! sSEND				--- update TXWRITE register and command WIZnet to SEND
	WAITSEND
  	txsize ~ 						--- reset txsize (buffer empty)
 	;
pub ?SEND							--- check to see if anything still needs to be sent
 	txsize @ 0EXIT						--- Exit if buffer empty
	LANSEND
	;
pub ?SENDPOLL
	txtime @ CNT@ - ABS CLKFREQ 7 SHR >			--- 1/128 sec timeout since last
 	0EXIT ?SEND
	;


--- send out a character through the WIZnet - either block mode or autosend
pub LANEMIT ( ch -- )
 	@txwr W@
	OVER SWAP ( ch ch index ) @TX LC!			\ write character to buffer
  	@txwr W++ CNT@ txtime !
 	txsize ++
	$0A = autosend C@ AND					\ SEND if this is an CRLF end of line and autosend is active
 	txsize @ $3F0 > OR 					\ AUTOSEND if buffer size is large enough already
	  IF LANSEND THEN
	;


pub LANKEY ( -- ch )
 	RXSIZE@
 	IF
 	  RXREAD LW@ DUP @RX LC@ 				--- read a character from the receive buffer
 	  SWAP 1+ RXREAD LW! sRECV				--- update read index
 	ELSE
 	  keypoll W@ ?DUP IF CALL THEN				--- implement a keypoll for when LANKEY is doing nothing
 	  0							--- return with a null character
 	THEN
	;

( Redirect console output to the LAN )
pub LAN
	' LANEMIT uemit W!
 	' LANKEY ukey W!
	@txwr @ -1 = IF TXWRITE LW@ @txwr ! THEN		--- force an update (assumed valid) if the index is "invalid"
	;

( diagnostic to the terminal to reflect what is begin typed to the LAN )
pub LANCONEMIT
 	DUP CONEMIT LANEMIT
 	;

( allow input from both the serial console and the LAN socket )
pub LANCONKEY ( -- ch )
 	LANKEY KEY OR
 	;

--- use both LAN and CON for output
pub LANCON
	CON CR LAN
 	' LANCONEMIT uemit W!
	;




--- set WIZnet chip as source for any DUMP type operations
pub WIZ    DUMP: LC@ LW@ L@ ;

$800	== WBUFSZ


pri LSEND 		DUP @txwr W+! TXWRITE LW@ SWAP @TX LWRITE LANSEND ;

LONG vread
LONG vwrite --- maintain a read/write pointer for any virtual memory writes


---  Read from WIZnet buffer into file until buffer is exhausted
--- Used by FTP STOR
pri LREADSKT ( dst --- )
	vwrite !
	BEGIN
	  RXSIZE@ ?DUP
	WHILE ( cnt )
	  RXREAD LW@ ( cnt index ) 			--- fetch receive read index for this socket
 	  SWAP
	  vwrite @ SWAP ( wizptr filedst cnt )
	  DUP >R
	  @RX LREAD vwrite ++ 				--- read WIZnet buffer directly into file and update vwrite+1
	  R> RXREAD LW@ + RXREAD LW! 			--- update read index in advance
	  sRECV 						--- signal that buffer up to READ index has been read
	  #10 ms
	REPEAT
	;


16 LONGS sktbuf						--- holding buffer for socket registers

pri @SKTBUF 		sktbuf + ;

pri .IPX	DUP >B $30A PRINTNUM 8 SHR ;

pub .IP1
	 @SKTBUF U@
	.IPX "." EMIT .IPX "." EMIT .IPX "." EMIT .IPX DROP
	;

pri .IP ( off -- )		DECIMAL 3 SPACES .IP1 ;



pri .PTR  		C@++ 8 SHL SWAP C@ + ( $7FF AND ) $2410 PRINTNUM "." EMIT ;

pri .@SKTBUF		@SKTBUF .PTR ;

IFNDEF SWAPB
pri SWAPB ( word -- word2 	\ Swap the bytes in a 16-bit word )
 	DUP 8 SHR SWAP >B 8 SHL OR
 	;
}

IFNDEF .TIME
pub .TIME	." 00:00:00" ;
}

{
SKT HH:MM:SS MODE  PORT DEST  TXRD TXWR RXRD RXWR RXSZ  IR STATUS        	IP ADDR
#1  00:00:00 TCP 	21 52775 967F.967F.	.	.	. 00 17 ESTABLISHED	192.168.016.002.
}
pub .SKTHD		CR BOLD PRINT" SKT HH:MM:SS MODE  PORT  DEST TXRD TXWR RXRD RXWR RXSZ  IR STATUS            IP ADDR" PLAIN ;

 pub .SKT
	DUP SOCKET 0 sktbuf $30 @SOCKET LREAD					--- read in registers
	0 @SKTBUF C@ IF CR "#" EMIT . 2 SPACES .TIME ELSE DROP BREAK
 	0  @SKTBUF C@ SPACE
	4* " CLSDTCP UDP IPRWMACRPPP !06!!07!" + 4 CTYPE SPACE
	4 @SKTBUF W@ SWAPB $250A PRINTNUM SPACE					--- PORT
 	16 @SKTBUF W@ SWAPB $250A PRINTNUM SPACE 				--- DPORT
	$22 .@SKTBUF $24 .@SKTBUF						--- TXRD TXWR
 	$28 .@SKTBUF $2A .@SKTBUF						--- RXRD RXWR
 	$26 .@SKTBUF SPACE							--- RX SIZE
  	2 @SKTBUF C@ .BYTE SPACE						--- INT REG
 	3 @SKTBUF C@ DUP .BYTE							--- STATUS REG
\	pri .SSTAT ( byte -- )							--- STATUS DESCRIPTION
 	SWITCH SPACE
 	  0 CASE PRINT" closed      " BREAK
 	$13 CASE PRINT" INIT        " BREAK
 	$14 CASE PRINT" LISTEN      " BREAK
	$16 CASE PRINT" SYNRECV     "  $0C .IP BREAK
   	$17 CASE PRINT" ESTABLISHED "  $0C .IP BREAK
  	$18 CASE PRINT" FIN WAIT    "  $0C .IP BREAK
 	$1C CASE PRINT" closed wait " $0C .IP BREAK
 	$22 CASE PRINT" UDP OPEN    " BREAK
 	$32 CASE PRINT" IPRAW OPEN  " BREAK
 	;

pub .SOCKET ( n -- )	.SKTHD .SKT ;

pub .SOCKETS
	.SKTHD 8 0 DO I .SKT LOOP CR
	;
pub ifconfig
	0 sktbuf $40 @COMMON LREAD	\ read in common registers
 	BOLD
	CR PRINT" NETWORK STATUS:" PLAIN
 	CR PRINT" LINK " 		1 $2E @SKTBUF SET? IF PRINT" *UP*" ELSE PRINT" DOWN" THEN
	CR PRINT" HARDWARE: " PCB$ PRINT$ PRINT"  using WIZnet W5500 V" $39 @SKTBUF C@ .
 	CR PRINT" SRC IP "	$0F .IP
 	CR PRINT" MASK   "	5 .IP
 	CR PRINT" GATEWAY"	1 .IP
	CR PRINT" MAC       "	9 @SKTBUF 6 ADO I C@ .BYTE PRINT" ." LOOP
	.SOCKETS
	;

\ SOCKETS - W5500 has 8 sockets - define 4 of them
0	== NETMAN --- network management
1	== FTP
2	== FTPDAT
3	== TELNET
4	== HTTP --- uses sockets 4..7
" NETFTPDATTELWEBWEBWEBWEB" 0 STRING skt$



{ NOTE: Use WIZ DUMP or variation thereof instead of WDUMP

0 40 @COMMON WIZ DUMP

0 40 1 SKT @TX DUMP

}


IFDEF EASYNET.fth
EASYNET.dat DATORG
FORGET EASYNET.fth
}

pub EASYNET.fth		PRINT" WIZNET NETWORK SERVERS 160707.1500 " ;

DATPTR	== EASYNET.dat




pub READYLED		DROP ;
pub LANLED		DROP ;
pri LEDS		NOP ;

pri UNKNOWN
 	delim C@ 1+ $0D <> IF BEGIN KEY $0D <> WHILE DROP REPEAT THEN		\ if there is more than get it and discard
	LANCON PRINT" 550 Unknown command " @WORD PRINT$ CR LANSEND
	;

WORD un
IFNDEF QUIET
\ Turn the Forth console into a quiet non-echoing/prompting command interpreter
pub QUIET
	IF
 	  OFF ECHO OFF OK				--- Non-interactive mode - just accept "commands"
	  ' NOP 					--- Disable auto space (added when the console processes a word)
	  ' UNKNOWN 					--- Respond with negative FTP style message for unknown
 	ELSE
 	  ON ECHO ON OK
 	  0 0						--- reset unum and prompt - default shell processing
 	THEN
 	unum W! DUP prompt W! prompt 2+ W!				--- set vector for further processing if not found or not number
	;
}


\ LANLED blink variables
BYTE ledcnt
BYTE ledon


( Console diagnostic message handler )
BYTE msgs
pub MSGS ( on/off -- ) msgs C! ;

LONG msgstk
pri <MSG ( level -- )				--- print message to console if message priority is => message setting with 0 being the highest
	uemit @ msgstk ! msgs C@ <=
 	  IF CON ELSE NULLOUT THEN
 	<CR> REVERSE .DT PRINT"  #" SKT@ DUP PRINT SPACE 3 * skt$ + 3 CTYPE SPACE
 	;
pri MSG>	PLAIN CR msgstk @ uemit ! ;


pri FlushSkt	 	  @WORD 1- BL ERASE ;		--- erase any garbage that has already been accumulated


{HELP LANSKT ( -- )
Set console's socket backup to the current one - allows console processing of server commands
}
pub LANSKT		SKT@ skt2 C! LAN ;

{HELP UpdateTXWR
Due to limitations of WIZnet read/write register access a copy is maintained and updated when possible
}
pri UpdateTXWR
	TXWRITE LW@ @txwr W! 						--- TXWRITE is now readable - buffer it
	;


TIMER contd

pri KEEPALIVE		#300,000 contd TIMEOUT ;	   			--- ms 300 seconds = 5 MINS


BYTE disreq									--- background timer cog can only request a disconnect
pri DISCREQ		disreq C~~ ;						--- timeout sets disreq flag which is handled by main loop

pub CONNECTED? ( -- flg )
	sCONNECTED? 								--- Has CONNECT interrupt been set?
	DUP IF 								--- and save result
	  OFF ECHO 								--- Setup Tachyon to handle command/response mode
         UpdateTXWR
	  ( 1 <MSG PRINT" CONNECT " MSG> )
	  CON SKT@ .SOCKET 							--- let console know what is happening
 	    BEGIN BEGIN KEY 0= UNTIL 3 ms KEY 0= UNTIL			--- discard any console input as well
 	  LAN
	  KEEPALIVE
 	  ON LANLED
	  FlushSkt								--- flush out anything already sitting there
	  ' DISCREQ contd ALARM
	THEN
	;
{HELP DISCONNECT
Disconnect socket (which may progress through DISCON WAIT) and flip the shell back to the console
Report the socket status
}
pub DISCONNECT	sDISCON CONIO SKT@ .SOCKET LAN ;

pub DSCTD? ( -- flg )
	disreq C@ sESTAB? AND
 	  IF DISCONNECT disreq C~ 0 contd ALARM THEN 			--- process disconnect req if active

 	sCLOSED? sCLOSING? OR 						--- or if it's closed OR closing or $1C?
 	sSTAT@ $1C = OR						--- CLOSE WAIT - check this out - gets stuck on this
( #2  22:09:58 TCP 	80 39416 EE08.EE08.  83.  83.	. 00 1C closed wait	150.070XXXX )
 	sDISCON? OR DUP 							--- disconnect interrupt? OR closed - ok
 	IF
	( 1 <MSG PRINT"   LISTENING " MSG> )
	  sCLOSE sOPEN sLISTEN
 	  OFF QUIET
	  OFF LANLED
 	  FlushSkt CONIO
 	  SKT@ .SOCKET
	THEN
	;
{
pri DISCONNECTED? ( -- flg )
	disreq C@ sESTAB? AND
 	  IF DISCONNECT disreq C~ 0 contd ALARM THEN 			--- process disconnect req if active

 	sCLOSED? sCLOSING? OR 						--- or if it's closed OR closing or $1C?
 	sSTAT@ $1C = OR						--- CLOSE WAIT - check this out - gets stuck on this
( #2  22:09:58 TCP 	80 39416 EE08.EE08.  83.  83.	. 00 1C closed wait	150.070XXXX )
 	sDISCON? OR DUP 							--- disconnect interrupt? OR closed - ok
 	IF
	( 1 <MSG PRINT"   LISTENING " MSG> )
	  sCLOSE sOPEN sLISTEN
 	  OFF QUIET
	  OFF LANLED
 	  FlushSkt CONIO
 	  SKT@ .SOCKET
	THEN
	;
}
pri CONNECT
	( 1 <MSG PRINT" CONNECTING " )
	#3000 contd TIMEOUT							--- allow up to 3 seconds to connect
	  BEGIN sESTAB? contd TIMEOUT? OR UNTIL 				--- Wait for it to be established (WIZNET regsiter quirks)
 	KEEPALIVE
	TXWRITE LW@ @txwr W! 						--- refresh txwrite (treat as new connection)
	( PRINT" ---  CONNECTED " MSG> )
	;

BYTE constat 									--- relay connection status to application

{ TELNET IACs
pri /DO		#253
pri IAC ( cmd -- ) #255 EMIT EMIT ;
pri /SB		#250 IAC ;
pri /SE		#240 IAC ;
pri /WILL		#251 IAC ;
pri /WONT		#252 IAC ;
pri /DONT		#254 IAC ;

pri /NAWS		#31 EMIT ;
pri /LINEMODE 	#34 EMIT ;
pri /ECHO		1 EMIT ;
}


--- define a "bye" especially for telnet sessions.

ALIAS DISCONNECT BYE
ALIAS DISCONNECT QUIT

pri ?TELNET
---	check telnet socket new connection?
	TELNET SOCKET CONNECTED?
	IF
---	indicate connection status active as Telnet
 	  "T" constat C!
---	be interactive but not reset etc with certain controls
  	  1 flags 1+ SET
	  ledcnt C~
\ 	  CON CR LAN
---		let 1/128 timeout handle characters or blocks
	  LAN  autosend C~
---	execute IAC script to setup remote telnet client
	  " TELNET.INI" FOPEN$ IF FILEIN THEN
 	  WAITSEND 	  " WELCOME.TEL" FOPEN$
 	    IF (cat)
---	default Welcome banner
 	    ELSE PRINT" WELCOME TO THE TACHYON WIZNET TELNET SESSION!" CR
 	    THEN
---	Reset rx buffer and receive --- redirect console to this LAN socket
	  FlushSkt LANSKT
 	  OFF QUIET OFF ECHO
	  KEEPALIVE
	THEN
---	Process disconnection
	DSCTD? IF
 	  ( 0 <MSG PRINT" SESSION ENDED " CR MSG> )
 	  "t" constat C! CONIO
 	  1 flags 1+ CLR
 	THEN
	;


pub RESOLVE ( namestr -- ip )
 	HTTP SOCKET
 	;
{
HTTP/1.0 302 Found
Cache-Control: private
Content-Type: text/html; charset=UTF-8
Location: http://www.google.com.au/?gfe_rd=cr&ei=UH_EVO24O63u8wfd74GYAQ
Content-Length: 262
Date: Sun, 25 Jan 2015 05:29:52 GMT
Server: GFE/2.0
Alternate-Protocol: 80:quic,p=0.02

<HTML><HEAD><meta http-equiv="content-type" content="text/html;charset=utf-8">
<TITLE>302 Moved</TITLE></HEAD><BODY>
<H1>302 Moved</H1>
The document has moved
<A HREF="http://www.google.com.au/?gfe_rd=cr&amp;ei=UH_EVO24O63u8wfd74GYAQ">here</A>.
</BODY></HTML>
}
pub GETTIME
 	HTTP SOCKET
 	" google.com" RESOLVE HTTP SOCKET $0C L!			--- contact google.com and set dest IP
 	LAN PRINT" GET /" CR						--- issue GET request

 	--- Date: Sun, 25 Jan 2015 05:29:52 GMT


 	;


\ Print the byte as a decimal number
pri .BYTEDEC ( byte -- )		>B $0A PRINTNUM ;

pri COMMA			"," EMIT  ;

pri .IPD ( long -- ) 	DUP #24 SHR .BYTEDEC COMMA DUP 16 SHR .BYTEDEC COMMA DUP 8 SHR .BYTEDEC COMMA .BYTEDEC ;


" user" BL STRING user$

" pass" BL STRING pass$

WORD dataport

#20 dataport W!			\ Default FTP data port


pri GETFNAME
	GETWORD DUP C@ "/" = IF 1+ THEN						--- adjust name if / is used
 	DUP LEN$ 1- OVER  + C@ "." = IF 0 OVER DUP LEN$ 1- + C! THEN 		--- remove final period
 	;


pri ECHOREQ
 	CON CR BEGIN LANKEY DUP EMIT 0= UNTIL					--- Just echo the request to the console for now
 	;

( FTP COMMANDS )
( FTP COMMANDS LINK )


{HELP USER
FTP COMMAND
Syntax: USER username
Send this command to begin the login process. username should be a valid username on the system, or "anonymous" to initiate an anonymous login.
}
pre USER
	ON LANLED
	GETWORD user$ $!
	PRINT" 331 User admin OK. Password required" CR
	;

pre PASS
	ON LANLED
	GETWORD pass$ $!
	PRINT" 230 OK. Current restricted directory is /" CR
	;

LONG type

{HELP TYPE
FTP COMMAND
Syntax: TYPE 	type-character [second-type-character]
Sets the type of file to be transferred. type-character can be any of:
A - ASCII text
E - EBCDIC text
I - image (binary data)
L - local format
For A and E, the second-type-character specifies how the text should be interpreted. It can be:
N - Non-print (not destined for printing). This is the default if 	second-type-character is omitted.
T - Telnet format control (<CR>, 	<FF>, etc.)
C - ASA Carriage Control
For L, the second-type-character specifies the number of bits per byte on the local system, and may not be omitted.
}
pre TYPE
	ON LANLED
	GETWORD type 4 CMOVE
	LAN PRINT" 200 TYPE is now " type C@ EMIT CR
	;

{HELP PORT
FTP COMMAND
Syntax: PORT 	a1,a2,a3,a4,p1,p2
Specifies the host and port to which the server should connect for the next file transfer.
This is interpreted as IP address a1.a2.a3.a4, port p1*256+p2.
}
pre PORT 	\ accept port number
	ON LANLED
	GETWORD NUMBER dataport W!
	LAN PRINT" 200 Port is now " dataport W@ .DEC CR
	;

\ Define min and max for FTP passive ports
#40,000 	== #ftpmin
#50,000 	== #ftpmax

\ 227 Entering Passive Mode (192,168,16,106,248,252)
{HELP PASV
FTP COMMAND
Syntax: PASV
Tells the server to enter "passive mode". In passive mode, the server will wait for the client to establish a connection with it rather than attempting to connect to a client-specified port. The server will respond with the address of the port it is listening on, with a message like:
227 Entering Passive Mode 	(a1,a2,a3,a4,p1,p2)
where a1.a2.a3.a4 is the IP address and p1*256+p2 is the port number.
}
pre PASV
	ON LANLED
	#ftpmax #ftpmin GETRND dataport W!						\ pick a random port in the specified range
	FTPDAT SOCKET sCLOSE 								\ Prep the data port socket
	TCP dataport W@ SetPORT sOPEN sLISTEN   					\ Set the port and open listen for connection
 	( 1 <MSG PRINT" FTPDAT CONNECT " MSG> )
	\ respond that all is accepted
	FTP SOCKET LAN									\ switch back to FTP socket
	PRINT" 227 Entering Passive Mode with port " \ dataport W@ .DEC		\ Response with msg and port
	PRINT" (" @sip EE@ .IPD COMMA dataport 1+ C@ .BYTEDEC COMMA dataport C@ .BYTEDEC PRINT" )" CR
	;

{HELP SYST
FTP COMMAND
Syntax: SYST
Returns a word identifying the system, the word "Type:", and the default transfer type (as would be set by the TYPE command).
For example: UNIX Type: L8
}
pre SYST
	ON LANLED
	PRINT" 215 Unix Type: L8" CR
	;




pre FEAT
	ON LANLED
	PRINT" 211 no Features supported" CR
{	PRINT" 211-no Features supported" CR
 	PRINT"  yet to be implemented" CR
 	PRINT" 211 End" CR
}
	;

{HELP MDTM
FTP COMMAND
  The server-PI will respond to the MDTM command with a 213 reply
   giving the last modification time of the file whose pathname was
   supplied, or a 550 reply if the file does not exist, the modification
   time is unavailable, or some other error has occurred.

      mdtm-response = "213" SP time-val CRLF /
                      error-response
Example response:
213 19980615100045.014
}
pre MDTM
 	ON LANLED
  	GETFNAME FOPEN$ DROP
 	autosend C~~
	PRINT" 550 Modification time not available" CR
	;


{HELP CDUP
FTP COMMAND
Syntax: CDUP
Makes the parent of the current directory be the current directory.
}
pre CDUP
	  ON LANLED
	PRINT" 250 Directory successfully changed" CR
	;


" FILENAME.TXT " 0 STRING cwd$

{HELP CWD
Syntax: CWD remote-directory
Makes the given directory be the current directory on the remote host
}
pre CWD ( <name> )
	ON LANLED
	GETWORD DUP cwd$ $! DUP " /" $= SWAP LEN$ 0= OR
 	IF
 	  PRINT" 250 okay"
	ELSE
	  PRINT" 550 Not a directory"
	THEN
 	CR
	;

{HELP PWD
Syntax: PWD
Returns the name of the current directory on the remote host.
}
pre PWD
	ON LANLED
	PRINT" 257 " $22 EMIT "/" EMIT $22 EMIT PRINT"  is your current location" CR
	;



pub ?DISCONNECT   	#300 ms DISCONNECT ;

{HELP LIST
FTP COMMAND
Syntax: LIST [remote-filespec]
If remote-filespec refers to a file, sends information about that file. If remote-filespec refers to a directory, sends information about each file in that directory. remote-filespec defaults to the current directory. This command must be preceded by a PORT or PASV command.
}
pre LIST
 	delim 1+ C@ BL = IF GETWORD DROP THEN                 --- ignore a remote-filespec
 	uemit W@ --- allow this to be dumped to the console in interactive mode
 	IF
	  KEEPALIVE
	  ON LANLED LANCON	PRINT" 150 Here comes the directory listing" CR LANSEND
	  ( 1 <MSG PRINT" LIST REQUEST " MSG> )
	  FTPDAT SOCKET CONNECT
 	  autosend C~
	  LAN .LIST							--- Send off the directory listing in compatible format
	  LANSEND WAITSEND
 	  #50 ms
	  FTPDAT SOCKET DISCONNECT
	  autosend C~~
	  FTP SOCKET LANCON PRINT" 226 Directory send OK" CR LANSEND
	  ( 1 <MSG PRINT" LIST SENT " MSG> )
	  ?DISCONNECT
 	ELSE
	  CR .LIST
	THEN
	;

{ This is how most UNIX, Novell, and MacOS ftp servers send their time
                                  Jul 06 12:57 or Jul  6  1999
-rwxrwxrwx	1 502  	500 	674 Sep  4  2014 HELP.TXT
-rwxrwxrwx	1 502  	500   65536 Sep  4  2014 FIRMWARE.ROM
}
{ Feature list from NAS FTP
211-Extensions supported:
 EPRT
 IDLE
 MDTM
 SIZE
 REST STREAM
 MLST type*;size*;sizd*;modify*;UNIX.mode*;UNIX.uid*;UNIX.gid*;unique*;
 MLSD
 TVFS
 ESTP
 PASV
 EPSV
 SPSV
 ESTA
 AUTH TLS
 PBSZ
 PROT
 UTF8
211 End.
}


WORD blkpoll
$200 == BUFSIZ

--- BLKSEND will send a file sector by sector
pri BLKSEND ( xaddr cnt -- )							\ send chunks of up to one complete sector at buffer address
	?DUP 0= IF DROP EXIT THEN
	BEGIN
	  OVER FSADR OVER BUFSIZ MIN ( xaddr cnt bufadr bufcnt )		\ grab it a sector at a time
	  >L 										\ save the bufcnt (referenced as IX)
	  IX LSEND									\ copy source buffer directly to socket tx buffer
         blkpoll W@ ?DUP IF CALL THEN						\ callback hook used by applications
	  ( src cnt )
	  IX - SWAP L> + SWAP ( xaddr cnt )					\ update source parameters block by block
	  DUP 0=									\ until the source buffer is exhausted (cnt=0)
 	  sCLOSED? OR									\ or if closed
	UNTIL
	2DROP
	;

--- Send the currently open file in block mode
pub SENDFILE ( offset -- )	FSIZE@ OVER - BLKSEND ;

{HELP SIZE
FTP COMMAND
Syntax: SIZE remote-filename
Returns the size of the remote file as a decimal number.
}
pre SIZE
	ON LANLED
  	GETFNAME FOPEN$
 	( 1 <MSG PRINT" SIZE " FILE$ PRINT$ MSG> )
 	LANCON
	IF
	  PRINT" 213 "  FSIZE@ #10 PRINTNUM
	ELSE
 	  PRINT" 550 Could not get file size."
 	THEN
 	CR LANSEND
	;
{

0000_C800:   55 53 45 52  20 61 6E 6F   6E 79 6D 6F  75 73 0D 0A   USER anonymous..
0000_C810:   50 41 53 53  20 63 68 72   6F 6D 65 40  65 78 61 6D   PASS chrome@exam
0000_C820:   70 6C 65 2E  63 6F 6D 0D   0A 53 59 53  54 0D 0A 50   ple.com..SYST..P
0000_C830:   57 44 0D 0A  54 59 50 45   20 49 0D 0A  50 41 53 56   WD..TYPE I..PASV
0000_C840:   0D 0A 53 49  5A 45 20 2F   0D 0A 50 41  53 56 0D 0A   ..SIZE /..PASV..
0000_C850:   43 57 44 20  2F 0D 0A 4C   49 53 54 20  2D 6C 0D 0A   CWD /..LIST -l..

550 Could not get the file size
}

--- send accepted or rejected message to FTP client
pri FTPMSG ( flg -- )
 	FTP SOCKET LANCON
	IF
	  ON LANLED
	  PRINT" 150 Accepted data connection for " FILE$ PRINT$ CR LANSEND
	  FTPDAT SOCKET CONNECT LAN
 	ELSE
	  PRINT" 550 File not available" CR LANSEND
	THEN
 	;

--- FTP rename file request ---

" FILENAME.EXT" 0 STRING RNFR$   \ place for source string for rename (PBJ: 0 STRING or #12 > req for null term.)

{HELP RNFR
FTP COMMAND
Syntax: RNFR from-filename
Used when renaming a file. Use this command to specify the file to be renamed; follow it with an RNTO command to specify the new name for the file.
}
pre RNFR
 	GETFNAME RNFR$ $!
 	FTP SOCKET LANCON PRINT" 350 Waiting for RNTO" CR LANSEND
 	;

{HELP RNTO
FTP COMMAND
Syntax: RNTO to-filename
Used when renaming a file. After sending an RNFR command to specify the file to rename, send this command to specify the new name for the file.
}
pre RNTO
 	RNFR$ FOPEN$ DROP GETFNAME RENAME$
 	FTP SOCKET LANCON PRINT" 250 Rename done" CR LANSEND
 	;


{
Command:	RETR PREVIOUS.ROM
Response:	150 Accepted data connection for PREVIOUS.ROM
Response:	226 File successfully transferred
Status:	File transfer successful, transferred 65,536 bytes in 1 second
}

--- FTP Retrieve a file i.e. RETR /LOG0001.TXT ---
{HELP RETR
FTP COMMAND
Syntax: RETR remote-filename
Begins transmission of a file from the remote host. Must be preceded by either a PORT command or a PASV command to indicate where the server should send data.
}
pri (RETR) ( position flg/addr -- )
 	DUP FTPMSG
 	IF ( position )
	  FILE$ 3 RIGHT$ " LOG" $=
  	  IF										// If it's a log file then just send up to EOF marker
 	    DROP 0 APPEND
 	      IF @FWRITE @ OVER - 16 MAX ELSE FSIZE@ THEN
 	    BLKSEND 									// Just send all the text up to the EOF or at least 16
 	  ELSE 									// else send the whole file
 	    SENDFILE
 	  THEN
	  #100 ms
	  ?SEND
  	  DISCONNECT --- close the data connection
	  FTP SOCKET
 	  LANCON PRINT" 226 File successfully transferred" CR LANSEND 	// Announce successful transfer
 	ELSE
 	  DROP
	THEN
 	FTP SOCKET ?DISCONNECT
	;

pre RETR
	KEEPALIVE
	GETFNAME FOPEN$ 								// get the file name and try to open it
 	0 SWAP (RETR)
	;


{HELP STOR
FTP COMMAND
Syntax: STOR remote-filename
Begins transmission of a file to the remote site. Must be preceded by either a PORT command or a PASV command so the server knows where to accept data from.
Usage
Command:	STOR PREVIOUS.ROM
Response:	150 Accepted data connection for PREVIOUS.ROM
Response:	250 File rcvd PREVIOUS.ROM
Status:	File transfer successful, transferred 65,536 bytes in 5 seconds
}
pre STOR
	KEEPALIVE
	GETFNAME 							--- get file name to store
	FOPEN$ RW 							--- try to open it for overwrite
	DUP FTPMSG 							--- send appropriate FTP message if accepted or not, connect to FTPDAT if accepted
 	?SEND
 	IF
 	  ( 1 <MSG PRINT" STOR " FILE$ PRINT$ MSG> )
	  FTPDAT SOCKET
	  BEGIN
	    @FWRITE @
	    LREADSKT ( dst --- )
 	    vwrite @ @FWRITE !							--- update file write index
 	    DSCTD?
 	  UNTIL
	  FTP SOCKET LANCON PRINT" 250 File rcvd " FILE$ PRINT$ CR LANSEND
 	  FSTAMP
 	  ( 1 <MSG PRINT" FILE RCVD"  MSG>)
	THEN
    	FTPDAT SOCKET ?DISCONNECT
 	FTP SOCKET ?DISCONNECT
	;




{HELP REST
FTP COMMAND
Syntax: REST position
Sets the point at which a file transfer should start; useful for resuming interrupted transfers. For nonstructured files, this is simply a decimal number. This command must immediately precede a data transfer command (RETR or STOR only); i.e. it must come after any PORT or PASV command.
}
pre REST ( <position> )
	GETWORD NUMBER 				--- read the offset specified
 	(RETR)						--- retrieve as usual from this offset
	;




{                   Tachyon FTP session
Status:	Connecting to 192.168.16.151:21...
Status:	Connection established, waiting for welcome message...
Response:	220 WELCOME TO THE TACHYON WIZNET FTP SESSION!
Command:	USER admin
Response:	331 User admin OK. Password required
Command:	PASS *****
Response:	230 OK. Current restricted directory is /
Status:	Server does not support non-ASCII characters.
Status:	Connected
Status:	Retrieving directory listing...
Command:	CWD /
Response:	250 Directory successfully changed
Command:	TYPE I
Response:	200 TYPE is now I
Command:	PASV
Response:	227 Entering Passive Mode with port (192,168,16,151,173,118)
Command:	LIST
Response:	150 Here comes the directory listing
Response:	226 Directory send OK
Status:	Directory listing successful

}



( *** FTP SERVER 	- relies on the Forth console to interpret FTP commands directly *** )

pri ?FTP
	FTP SOCKET
	CONNECTED? 						--- examine interrupt register for a new connection etc
	IF
 	  "F" constat C!
\	  CON CR LAN
	  LAN ON QUIET
{
	  " WELCOME.FTP" FOPEN$
	  IF
 	    LAN (cat)
 	  ELSE
}
 	autosend C~~

	    LANCON PRINT" 220 WELCOME TO THE TACHYON WIZNET FTP SESSION!" CR --- Welcome banner
\  	  THEN
	  KEEPALIVE 						--- Give FTP a maximum 5 min session
	  LANSKT 						--- makes sure the console uses this connection and socket
	THEN
	DSCTD? IF "f" constat C! CONIO THEN
	;



( HTTP COMMANDS )

--- some WIP here while I sort out my webpage files and content formatting etc


\ Sample content header - just for testing
pri CONTENT ( str -- )
	1 <MSG PRINT" CONTENT = " DUP PRINT$  MSG>
	LANCON autosend C~
	PRINT" HTTP/1.1 200 OK" CR
	PRINT" Date: Tue, 03 Dec 2013 04:19:05 GMT" CR
	PRINT" Server: Tachyon Forth " CR
	PRINT" Last Modified: Tue, 01 Jan 2014 12:00:00 GMT" CR
	PRINT" Accept-Ranges: bytes" CR
	PRINT" Content-Length: " FSIZE@ .DEC CR
	PRINT" Connection: close" CR
	PRINT" Content-Type: " PRINT$ CR
	autosend C~~ CR
	;








#48 BYTES GET$

#48 	== getsz	--- limited in size by GETWORD = 39 or implement new GETPARS to place string directly into GET$

pri ?CONTENT     \  MJB maybe later move to a table based approach
	 \ GET$ 3 RIGHT$
	"." GET$ LOCATE$ 1+   \ MJB prepare for other extensions like "JS"

	DUP " JPG" COMPARE$ IF DROP " image/jpg" CONTENT EXIT THEN		\ Send header for JPG files
	DUP " ICO" COMPARE$ OVER " PNG" COMPARE$ OR
	  IF DROP " image/png" CONTENT EXIT THEN					\ Send header for ICO/PNG files
	DUP " HTM" COMPARE$ IF DROP " text/html" CONTENT EXIT THEN		\ Send header for TEXT/HTML files
	DROP
	;

IFNDEF >UPPER
pri >UPPER  ( str1 --  ) --- Convert lower-case letters to upper-case
	DUP LEN$ ADO I C@ "a" "z" WITHIN IF I C@ BL XOR I C! THEN LOOP
	;
}

pri GETTXT
 	0 APPEND DROP @FWRITE @ OVER - 16 MAX BLKSEND			\ Just send all the text up to the EOF or at least 16
 	;
pri GETHTX --- temp def
	GETTXT
	;



--- Send the requested page to the web client ---
pri GETPAGE
	GET$ " /" COMPARE$ 								\ null GET / (root dir)
	  IF " HOME.HTM" DUP GET$ $!						\ Default request - open HOME.HTM
	  ELSE GET$ 1+ DUP >UPPER 							\ otherwise convert requested file name to uppercase
	  THEN
 	FOPEN$ NOT IF " HTTP404.HTM" FOPEN$ DROP THEN 				\ on file not found - use default 404 file
	LAN										\ Direct all output to the selected socket
	?CONTENT									\ Handle content headers
	1 <MSG PRINT"  --> " FILE$ PRINT$ MSG>				\ echo name of actual file served
	UpdateTXWR @txwr W@ TXREAD LW!

	FILE$ 3 RIGHT$
  	DUP " TXT" $= IF DROP GETTXT ?SEND EXIT THEN				\ plain text file  ( only uppercase ?? or is >upper used somewhere?)
 	" HTX" $= IF GETHTX ?SEND EXIT THEN 					\ we have a html template file HTX with embedded FORTH

	0 FSIZE@ BLKSEND								\ or just send the whole file
	?SEND
	;

{HELP HEAD
HTTP COMMAND
The HEAD method is identical to GET except that the server MUST NOT return a message-body in the response. The metainformation contained in the HTTP headers in response to a HEAD request SHOULD be identical to the information sent in response to a GET request. This method can be used for obtaining metainformation about the entity implied by the request without transferring the entity-body itself. This method is often used for testing hypertext links for validity, accessibility, and recent modification.
The response to a HEAD request MAY be cacheable in the sense that the information contained in the response MAY be used to update a previously cached entity from that resource. If the new field values indicate that the cached entity differs from the current entity (as would be indicated by a change in Content-Length, Content-MD5, ETag or Last-Modified), then the cache MUST treat the cache entry as stale.
}
pre HEAD  --- just repond back with the same head
	ECHOREQ
	;



{HELP GET
HTTP COMMAND
The GET method means retrieve whatever information (in the form of an entity) is identified by the Request-URI. If the Request-URI refers to a data-producing process, it is the produced data which shall be returned as the entity in the response and not the source text of the process, unless that text happens to be the output of the process.
The semantics of the GET method change to a "conditional GET" if the request message includes an If-Modified-Since, If-Unmodified-Since, If-Match, If-None-Match, or If-Range header field. A conditional GET method requests that the entity be transferred only under the circumstances described by the conditional header field(s). The conditional GET method is intended to reduce unnecessary network usage by allowing cached entities to be refreshed without requiring multiple requests or transferring data already held by the client.
The semantics of the GET method change to a "partial GET" if the request message includes a Range header field. A partial GET requests that only part of the entity be transferred, as described in section 14.35. The partial GET method is intended to reduce unnecessary network usage by allowing partially-retrieved entities to be completed without transferring data already held by the client.
The response to a GET request is cacheable if and only if it meets the requirements for HTTP caching described in section 13.
}
pre GET ( <name> -- \ Open up the file name and send it ) 		--- /index.htm HTTP/1.1
	KEEPALIVE
	LAN GETWORD DUP LEN$ getsz => IF DROP " GET$ to long" THEN
 	GET$ COPY$ 									--- get the name and store in GET$  \ GETWORD
	1 <MSG PRINT" GET " GET$ PRINT$ MSG>					--- Let me know about a request
	ECHOREQ
	GETPAGE
	;

1	== #hskts							--- select from 1 to 4 sockets for HTTP processing

{HELP ?HTTP
Service the HTTP server socket
}
pri ?HTTP
	HTTP #hskts ADO
 	I SOCKET
	CONNECTED?
	  IF
 	    "H" constat C!
	    ON QUIET 							--- Disable interactive mode prompts
	    KEEPALIVE
	    LANSKT 							--- Let the same socket talk to the foreground Forth console when it switches back
	  THEN
 	  DSCTD? IF "h" constat C! CONIO THEN
 	LOOP
	;


\ ----------------------------------------------------------------------------------------------------------- \

\ Check for control keys - this does not read the input buffer
pri ?CTRLS
	lastkey C@ SWITCH lastkey C~
 	^A CASE CONIO DISCARD OFF QUIET PRINT"   ENQ " PLAIN CONSOLE BREAK
	^W CASE [CON DISCARD ifconfig CON] BREAK
 	^K CASE keypoll W~ OFF QUIET CONIO BREAK 					--- kill background key poll (servers)
	^S CASE [CON DISCARD .SOCKETS CON] BREAK
	^Q CASE CON DISCARD 8 0 DO I SOCKET DISCONNECT LOOP BREAK
 	^C CASE CON OFF QUIET DEBUG BREAK
       SWITCH@ IF CONIO OFF QUIET PLAIN THEN
	;

pri ?LED
 	--- the LED should be blinking very briefly when it's idle but alive
	1 ledcnt C+! ledcnt C@ ledon C@ < LANLED LEDS
	--- reflect current connection status as a long or short blink
	constat C@ $61 < IF constat C@ ELSE 1 THEN ledon C!
 	;

BYTE fsave
BYTE netflgs
--- 1	inhibit console shortcuts

( Main server loop - checks and services sockets - 1.3ms when idle )
pub ?EASYNET
	?LED
	SKT@ skt2 C! 							--- swap current socket between foreground and background
 	FILE# fsave C! 0 FILE 						--- use foreground file
	?TELNET 								--- Poll the TELNET server
	?SENDPOLL
	?FTP 									--- Poll the FTP server
	?HTTP 									--- Poll the WEB server
	?SDCARD
\  	TRUE 8 0 DO I SOCKET sCLOSED? AND LOOP IF CONIO THEN			--- force console back to serial if not busy
 	skt2 C@ SOCKET 							--- restore foreground socket
 	fsave C@ FILE								--- restore foreground file
  	DEPTH 8 > IF !SP THEN							--- clean up abnormal stack
	1 netflgs SET? ?EXIT							--- skip controls if flag is set
\	?CTRLS									--- process console shortcuts
\	CONIO
	;

pub RESTART		CON PLAIN ifconfig CR CR REBOOT ;



pub EASYNET
	!SP
\	!PCB
	BOOT
	!WIZ #150 ms 							--- Setup I2CBUS LEDS (includes WIZNET RESET & PWRDN)
	#5000 @RTR LW! 16 @RCR LC!						--- setup retry counters
	CR PRINT" *** Tachyon Forth EASYNET Network Servers and EASYFILE File Server *** " CR CR
	RND fboot !
	MOUNT
	4 ledon C! 								--- just setup an LED blink time
	ON READYLED 								---  Now init the IP addresses (stored in high 64K EEPROM)
	TELNET SOCKET sCLOSE TCP #10001 PORT!  sOPEN sLISTEN 		--- Setup TELNET but on port 10001
	FTP SOCKET sCLOSE TCP #21 PORT! sOPEN sLISTEN 			--- Setup FTP
	HTTP #hskts ADO I SOCKET sCLOSE TCP #81 PORT! sOPEN sLISTEN LOOP
\	1 second
	PRINT"  ... ready! " CR
	ifconfig 									--- report WIZnet status
  	constat C~~
	' ?EASYNET keypoll W!  \ +POLL 						--- Poll the server in the background
	CR PRINT" * WEB, FTP, and TELNET servers running * "
	CR
 	;



END
